/* This file is autogenerated from the database. */
/* See src/desc/gen_proto */
/* Do not edit*/
entree functions_default[]={
{"TeXstyle",0,(void*)sd_TeXstyle,18,"",""},
{"breakloop",0,(void*)sd_breakloop,18,"",""},
{"colors",0,(void*)sd_colors,18,"",""},
{"compatible",0,(void*)sd_compatible,18,"",""},
{"datadir",0,(void*)sd_datadir,18,"",""},
{"debug",0,(void*)sd_debug,18,"",""},
{"debugfiles",0,(void*)sd_debugfiles,18,"",""},
{"debugmem",0,(void*)sd_debugmem,18,"",""},
{"echo",0,(void*)sd_echo,18,"",""},
{"factor_add_primes",0,(void*)sd_factor_add_primes,18,"",""},
{"factor_proven",0,(void*)sd_factor_proven,18,"",""},
{"format",0,(void*)sd_format,18,"",""},
{"graphcolormap",0,(void*)sd_graphcolormap,18,"",""},
{"graphcolors",0,(void*)sd_graphcolors,18,"",""},
{"help",0,(void*)sd_help,18,"",""},
{"histfile",0,(void*)sd_histfile,18,"",""},
{"histsize",0,(void*)sd_histsize,18,"",""},
{"lines",0,(void*)sd_lines,18,"",""},
{"linewrap",0,(void*)sd_linewrap,18,"",""},
{"log",0,(void*)sd_log,18,"",""},
{"logfile",0,(void*)sd_logfile,18,"",""},
{"nbthreads",0,(void*)sd_nbthreads,18,"",""},
{"new_galois_format",0,(void*)sd_new_galois_format,18,"",""},
{"output",0,(void*)sd_output,18,"",""},
{"parisize",0,(void*)sd_parisize,18,"",""},
{"parisizemax",0,(void*)sd_parisizemax,18,"",""},
{"path",0,(void*)sd_path,18,"",""},
{"prettyprinter",0,(void*)sd_prettyprinter,18,"",""},
{"primelimit",0,(void*)sd_primelimit,18,"",""},
{"prompt",0,(void*)sd_prompt,18,"",""},
{"prompt_cont",0,(void*)sd_prompt_cont,18,"",""},
{"psfile",0,(void*)sd_psfile,18,"",""},
{"readline",0,(void*)sd_readline,18,"",""},
{"realprecision",0,(void*)sd_realprecision,18,"",""},
{"recover",0,(void*)sd_recover,18,"",""},
{"secure",0,(void*)sd_secure,18,"",""},
{"seriesprecision",0,(void*)sd_seriesprecision,18,"",""},
{"simplify",0,(void*)sd_simplify,18,"",""},
{"sopath",0,(void*)sd_sopath,18,"",""},
{"strictargs",0,(void*)sd_strictargs,18,"",""},
{"strictmatch",0,(void*)sd_strictmatch,18,"",""},
{"threadsize",0,(void*)sd_threadsize,18,"",""},
{"threadsizemax",0,(void*)sd_threadsizemax,18,"",""},
{"timer",0,(void*)sd_timer,18,"",""},
{NULL,0,NULL,0,NULL,NULL} /* sentinel */
};
